//----------------------------------------------------------------------------------------------//
//						     STYLIZATION SUITE 2.0 FOR SILENT HORIZONS							//
//----------------------------------------------------------------------------------------------//
//																								//
//				     Stylization Suite Addon by LonelyKitsuune aka Skratzer						//
//						  for Skyrim SE ENB (DirectX 11 Shader Model 5)							//
//																								//
//			   Copyright (c) 2019-2020 LonelyKitsuune / T.Thanner - CC BY-NC-ND 4.0				//
//																								//
//----------------------------------------------------------------------------------------------//
//								==================================								//
//								//		Silent Horizons ENB     //								//
//								//								//								//
//								//		 by LonelyKitsuune      //								//
//								==================================								//
//----------------------------------------------------------------------------------------------//

#ifndef _STYLIZATION_SUITE_
#define _STYLIZATION_SUITE_

//----------------------------------------------------------------------------------------------//
//								       Internal UI Parameters									//
//----------------------------------------------------------------------------------------------//

UI_WHITESPACE(STYLE_WS1)
UI_WHITESPACE(STYLE_WS2)
UI_ELEMENT(_STYLE_Header,"     >>>>>>STYLIZATION SUITE<<<<<<")
UI_WHITESPACE(STYLE_WS3)
UI_ELEMENT(_OUTLINE_Header,              "|---------- CEL SHADING / OUTLINE")
bool   UISO_Enable        <string UIName="|- Outline - Enable";                                                              > = {false};
float  UISO_LineIntensity <string UIName="|- Outline - Intensity";                        float UIMin=0.0; float UIMax=20.0; > = {2.0};
float  UISO_LineCurve     <string UIName="|- Outline - Curve";                            float UIMin=0.7; float UIMax=10.0; > = {1.0};
float  UISO_DepthFade     <string UIName="|- Outline - Depth Fade";                       float UIMin=0.0; float UIMax= 1.0; > = {0.0};
float  UISO_LineThickness <string UIName="|- Outline - Line Thickness"; float UIStep=1.0; float UIMin=1.0; float UIMax=10.0; > = {1.0};
float3 UISO_LineColor     <string UIName="|- Outline - Color";                            string UIWidget="Color";           > = {0.0, 0.0, 0.0};
int    UISO_LineType      <string UIName="|- Outline - Type";                             int   UIMin=1;   int   UIMax=3;    > = {1};
bool   UISO_AgrZBufferFix <string UIName="|- Outline - Agressive Z-Buffer Fix";                                              > = {false};
bool   UISO_LimitToSkins  <string UIName="|- Outline - Limit to Skins";                                                      > = {false};


//----------------------------------------------------------------------------------------------//
//										      Structs											//
//																								//
//----------------------------------------------------------------------------------------------//

struct VertexShaderOutputStyle
{
   float4 pos       : SV_POSITION;
   float2 texcoord  : TEXCOORD0;
};


//----------------------------------------------------------------------------------------------//
//								     Functions and Variables									//
//																								//
//----------------------------------------------------------------------------------------------//

float ConvolvePixelMatrix(float3x3 PixelMatrix, float3x3 Kernel)
{
	float  Result  = dot(PixelMatrix[0], Kernel[0]);
		   Result += dot(PixelMatrix[1], Kernel[1]);
		   Result += dot(PixelMatrix[2], Kernel[2]);
	return Result;
}

float3x3 CreateLinDepthPixelMatrix(float2 Coords, float FarDepth, out float NonLinCenterDepth)
{
	float4 Samples[3];
	Samples[0]   = TextureDepth.Gather(Linear_Sampler, Coords - PixelSize * 0.5);
	Samples[1]   = TextureDepth.Gather(Linear_Sampler, Coords + PixelSize * 0.5);
	Samples[2].x = TextureDepth.SampleLevel(Point_Sampler, Coords + PixelSize * float2( 1.0,-1.0),0).x;
	Samples[2].y = TextureDepth.SampleLevel(Point_Sampler, Coords + PixelSize * float2(-1.0, 1.0),0).x;
	
	NonLinCenterDepth = Samples[0].y;
	
	Samples[0]    = FastLinDepth(Samples[0],    FarDepth);
	Samples[1]    = FastLinDepth(Samples[1],    FarDepth);
	Samples[2].xy = FastLinDepth(Samples[2].xy, FarDepth);
	
	float3x3 PixelMatrix = { Samples[0].w, Samples[0].z, Samples[2].x,
							 Samples[0].x, Samples[0].y, Samples[1].z,
							 Samples[2].y, Samples[1].x, Samples[1].y };
	
	return PixelMatrix;
}

float GetDepthDiscontinuities(float3x3 DepthMatrix, float Agressiveness)
{
	const float3x3 NeighbourAverage = { 0.125, 0.125, 0.125,
										0.125, 0.0,   0.125,
										0.125, 0.125, 0.125 };
	
	float NeiAverage = ConvolvePixelMatrix(DepthMatrix, NeighbourAverage);
	//float NeiMedian = CalculateMedian(DepthMatrix);
	
	float DepthDelta = DepthMatrix._22 - NeiAverage;
		  DepthDelta = saturate(1.0 - DepthDelta * Agressiveness);
	
	return DepthDelta;
}


//----------------------------------------------------------------------------------------------//
//										    Shaders												//
//																								//
//----------------------------------------------------------------------------------------------//

VertexShaderOutputStyle VS_DrawStyle(VertexShaderInput IN)
{
	VertexShaderOutputStyle OUT;
	OUT.pos      = float4(IN.pos.xyz, 1.0);
	OUT.texcoord = IN.txcoord.xy;
	
	return OUT;
}


float4 PS_StylePrepareEdges(VertexShaderOutputStyle IN) : SV_Target
{
	if(!UISO_Enable) discard;
	
	float SkinMask = saturate(1.0 - TextureColor.Sample(Point_Sampler, IN.texcoord).w);
	
	float    NonLinDepth;
	float3x3 DepthMatrix = CreateLinDepthPixelMatrix(IN.texcoord, 2000.0, NonLinDepth);
	float    LinDepth    = DepthMatrix._22;
	
	float FixAgressiveness = UISO_AgrZBufferFix ? 70.0 : 6.0;
	float ZBufferFix = GetDepthDiscontinuities(DepthMatrix, FixAgressiveness);
	
	float DepthFade = LinearStep(saturate(1.0 - UISO_DepthFade + DELTA),
								 saturate(0.8 - UISO_DepthFade), FastLinDepth(NonLinDepth, 200.0));
	
	return float4(LinDepth, SkinMask, DepthFade, ZBufferFix);
}


float4 PS_StyleDrawEdges(VertexShaderOutputStyle IN) : SV_Target
{
	float4 Color = TextureColor.Sample(Point_Sampler, IN.texcoord);
	
	if(!UISO_Enable) return Color;
	
	float  MaxDepthFadeMask = 0.0;
	float2 BaseWeight = (UISO_LineThickness + 1.0) * float2(2.0, 1.0);
	float  WeightSum  = 0.0;
	float2 Edges      = 0.0;
	
	[loop] for(float i=1.0; i <= UISO_LineThickness; i++)
	{
		float  Weight     = BaseWeight.x - i * 2.0;
		float2 Offsets[4] = { IN.texcoord.xyxy + PixelSize.xyxy * float4( i, 0,-i, 0),
							  IN.texcoord.xyxy + PixelSize.xyxy * float4( 0,-i, 0, i) };
		
		[unroll] for(int j=0; j<4; j++)
		{
			float4 CurrSample = RenderTargetRGBA64.SampleLevel(Linear_Sampler, Offsets[j],0);
			float  CurrWeight = Weight * CurrSample.w;
			
			MaxDepthFadeMask = max(MaxDepthFadeMask, CurrSample.z);
			
			Edges     += CurrSample.xy * CurrWeight;
			WeightSum += CurrWeight;
		}
	}
	
	[loop] for(float x=1.0; x <= UISO_LineThickness; x++)
	[loop] for(float y=1.0; y <= UISO_LineThickness; y++)
	{
		float  Weight     = BaseWeight.y - length(float2(x,y));
		float2 Offsets[4] = { IN.texcoord.xyxy + PixelSize.xyxy * float4(x, y,-x,-y),
							  IN.texcoord.xyxy + PixelSize.xyxy * float4(y,-x,-y, x) };
		
		[unroll] for(int j=0; j<4; j++)
		{
			float4 CurrSample = RenderTargetRGBA64.SampleLevel(Linear_Sampler, Offsets[j],0);
			float  CurrWeight = Weight * CurrSample.w;
			
			MaxDepthFadeMask = max(MaxDepthFadeMask, CurrSample.z);
			
			Edges     += CurrSample.xy * CurrWeight;
			WeightSum += CurrWeight;
		}
	}
	
	float4 CenterSample = RenderTargetRGBA64.Sample(Point_Sampler, IN.texcoord);
	
		  Edges     -= CenterSample.xy * WeightSum;
	float DepthEdge  = saturate(abs(Edges.y) + !UISO_LimitToSkins);
		  DepthEdge *= Edges.x * MaxDepthFadeMask;
		  DepthEdge *= UISO_LineIntensity * UISO_LineIntensity;
	
	
	[flatten] switch(UISO_LineType)
	{
		case 1: DepthEdge = saturate(abs(DepthEdge)); break;
		case 2: DepthEdge = saturate( DepthEdge);     break;
		case 3: DepthEdge = saturate(-DepthEdge);     break;
	}
	
	DepthEdge = pow(DepthEdge, UISO_LineCurve);
	Color.rgb = zerolim(Color.rgb - (UISO_LineColor * -2.0 + 1.0) * DepthEdge);
	
	return Color;
}



//--------------------------------- Stylization Technique Macro --------------------------------//

#define STYLE_TECHS(TECH_NAME, a,b) \
TECH11(COMBINE(TECH_NAME,a) <string RenderTarget="RenderTargetRGBA64";>, VS_DrawStyle(), PS_StylePrepareEdges())\
TECH11(COMBINE(TECH_NAME,b),                                             VS_DrawStyle(), PS_StyleDrawEdges())
#endif



